from pathlib import Path
from typing import Optional

from scfile import consts
from scfile import exceptions as exc
from scfile.file.base import FileDecoder, FileEncoder
from scfile.file.base.decoder import DATA, OPENER
from scfile.utils.types import PathLike


def is_supported(source: PathLike) -> bool:
    return Path(source).suffix.lstrip(".") in consts.SUPPORTED_SUFFIXES


def convert(
    source: PathLike,
    output: Optional[PathLike],
    decoder: type[FileDecoder[OPENER, DATA]],
    encoder: type[FileEncoder[DATA]],
    overwrite: bool = True,
):
    source = Path(source)
    output = Path(output or source.parent)

    new_suffix = encoder.suffix()
    new_source = source.with_suffix(f".{new_suffix}")

    # Check that file exists
    if not source.exists() or not source.is_file():
        raise exc.FileNotFound(source)

    # Full destination path
    destination = output / new_source.name

    # If overwrite is disabled
    # Validate that file with that name doesn't exist
    if not overwrite:
        counter = 1
        while Path(destination).exists():
            destination = output / Path(f"{new_source.stem} ({counter}).{new_suffix}")
            counter += 1

    # Create output directory
    destination.parent.mkdir(parents=True, exist_ok=True)

    # Convert and save file to destination path
    with decoder(source) as dec:
        dec.convert_to(encoder).save(destination)



import pathlib

import click

from scfile.consts import OutputFormats


PATH_TYPE = pathlib.Path

FILES = click.Path(path_type=PATH_TYPE, dir_okay=True, file_okay=True, exists=True, resolve_path=True)
OUTPUT = click.Path(path_type=PATH_TYPE, dir_okay=True, file_okay=False)
FORMATS = click.Choice(list(OutputFormats.MODELS))



import sys
import traceback
from typing import Any

import click
from rich import print

from scfile.enums import EchoPrefix as PREFIX


def excepthook(
    exc_type: type[BaseException],
    exc_value: BaseException,
    exc_traceback: Any,
):
    if isinstance(exc_value, click.ClickException):
        print(PREFIX.CLICK, exc_value)
    else:
        traceback.print_exception(exc_type, exc_value, exc_traceback)
        print()
        print(PREFIX.EXCEPTION, f"{exc_type.__name__} - {exc_value}.")

    sys.exit(1)



from .base import ScFileException
from .basic import FileBasicError


class FileDecodingError(FileBasicError):
    """Base exception occurring while file decoding."""

    pass


class FileParsingError(ScFileException):
    """Base exception occurring due to unexpected file structure."""

    pass


class FileUnsupportedError(ScFileException):
    """Base exception occurring intentionally for unsupported formats."""

    pass



from pathlib import Path
from typing import Optional, Sequence

from scfile import exceptions as exc
from scfile.enums import FileSuffix
from scfile.utils.types import PathLike

from . import formats as convert


DEFAULT_MODEL_FORMATS = (FileSuffix.OBJ,)

MODEL_CONVERTER_MAP = {
    FileSuffix.DAE: convert.mcsa_to_dae,
    FileSuffix.OBJ: convert.mcsa_to_obj,
    FileSuffix.MS3D: convert.mcsa_to_ms3d,
    FileSuffix.MS3D_ASCII: convert.mcsa_to_ms3d_ascii,
}


def auto(
    source: PathLike,
    output: Optional[PathLike] = None,
    formats: Optional[Sequence[FileSuffix]] = None,
    overwrite: bool = True,
    hdri: bool = False,
):
    """
    Automatically determines which format convert to.

    Args:
        source: Path to encrypted file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        formats (optional): Sequence of FileSuffix for preferred output models formats.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.
        hdri (optional): Convert ol texture as hdri sky. Defaults to False.

    Raises:
        FileSuffixUnsupported - if source suffix not in consts.SUPPORTED_SUFFIXES.

    Example:
        `auto("file.mcsa")`
    """

    if not formats:
        formats = DEFAULT_MODEL_FORMATS

    path = Path(source)
    suffix = path.suffix.lstrip(".")

    args = (source, output, overwrite)

    match suffix:
        case FileSuffix.MCSA | FileSuffix.MCVD:
            for fmt in formats:
                if converter := MODEL_CONVERTER_MAP.get(fmt):
                    converter(*args)

        case FileSuffix.MIC:
            convert.mic_to_png(*args)

        case FileSuffix.OL:
            if hdri:
                convert.ol_hdri_to_dds(*args)
                return

            convert.ol_to_dds(*args)

        case _:
            raise exc.FileSuffixUnsupported(path)



import sys
from pathlib import Path
from typing import Optional, Sequence

import click
from rich import print

from scfile.cli.excepthook import excepthook
from scfile.consts import CLI
from scfile.enums import EchoPrefix as PREFIX
from scfile.enums import FileSuffix
from scfile.exceptions import ScFileException
from scfile.utils import convert

from . import types


sys.excepthook = excepthook


@click.command(no_args_is_help=True, epilog=CLI.EPILOG)
@click.argument("FILES", type=types.FILES, nargs=-1, required=True)
@click.option("-F", "--formats", help="Preferred format for models.", multiple=True, type=types.FORMATS)
@click.option("-O", "--output", type=types.OUTPUT, help="Output results directory.")
@click.option("--subdir", is_flag=True, help="Recreate input subdirectories to output.")
@click.option("--hdri", is_flag=True, help="All ol textures in input files is hdri (skies).")
@click.option("--no-overwrite", is_flag=True, help="Do not overwrite file if already exists.")
@click.option("--silent", is_flag=True, help="Suppress all console echoes.")
def scfile(
    files: Sequence[Path],
    formats: Sequence[FileSuffix],
    output: Optional[Path] = None,
    subdir: bool = False,
    hdri: bool = False,
    no_overwrite: bool = False,
    silent: bool = False,
):
    # TODO: pretty click error handler and echo
    def echo(*args, **kwargs):
        if not silent:
            print(*args, **kwargs)

    # Reverse overwrite flag
    overwrite = not no_overwrite

    # Title echo
    echo(f"[b purple]SCFILE:[/] {CLI.VERSION}")
    echo()

    # Warn invalid flags
    if subdir and not output:
        echo(PREFIX.WARN, "[b]--subdir[/] flag cannot be used without specifying [b]--output[/] option.")
        echo()

    # Get root path for --subdir output
    first_path = files[0]
    root = first_path if first_path.is_dir() else None

    # Create glob from directory input
    if first_path.is_dir():
        # haven't figured out how to implement
        # multiple directories with --output and --recusive support
        if len(files) != 1:
            echo(PREFIX.ERROR, "Only one directory suported at once.")
            return

        files = tuple(first_path.rglob("**/*"))

    # Exclude directories and unsupported files
    files = tuple(filter(lambda path: path.is_file() and convert.is_supported(path), files))

    # Check that files not empty
    if not files:
        echo(PREFIX.ERROR, "No supported files found in provided arguments.")
        echo(PREFIX.INFO, CLI.FORMATS)
        return

    # Convert all files
    for source in files:
        destination = output

        # Get relative destination path
        if output and subdir and root:
            relative = source.parent.relative_to(root)
            destination = Path(output or "") / relative

        # Convert source file
        try:
            convert.auto(source, destination, formats, overwrite, hdri)
            echo(PREFIX.INFO, f"File '{source.name}' converted to '{destination or source.parent}'.")

        except ScFileException as err:
            echo(PREFIX.ERROR, str(err))

        except Exception as err:
            echo(PREFIX.EXCEPTION, f"'{source}' - {err}")



from .auto import auto
from .convert import convert, is_supported
from .formats import (
    mcsa_to_dae,
    mcsa_to_ms3d,
    mcsa_to_ms3d_ascii,
    mcsa_to_obj,
    mic_to_png,
    ol_hdri_to_dds,
    ol_to_dds,
)


__all__ = (
    "auto",
    "convert",
    "is_supported",
    "mcsa_to_dae",
    "mcsa_to_ms3d",
    "mcsa_to_ms3d_ascii",
    "mcsa_to_obj",
    "mic_to_png",
    "ol_to_dds",
    "ol_hdri_to_dds",
)



import io
import pathlib

from scfile.exceptions import FileStructureInvalid

from .base import BinaryIO


class BinaryBytesIO(io.BytesIO, BinaryIO):
    pass


class BinaryFileIO(io.FileIO, BinaryIO):  # type: ignore
    @property
    def path(self) -> pathlib.Path:
        return pathlib.Path(str(self.name))

    def validate_buffer_size(self, size: int):
        current_pos = self.tell()
        self.seek(0, io.SEEK_END)
        file_size = self.tell()

        # return pointer
        self.seek(current_pos)

        remaining = file_size - current_pos

        if remaining < size:
            raise FileStructureInvalid(self.path, current_pos)



import click
from rich import print

from scfile.cli.scfile import scfile


def main():
    try:
        scfile(standalone_mode=False)

    except (KeyboardInterrupt, click.exceptions.Abort):
        print("[b yellow]Operation Aborted.[/]")


if __name__ == "__main__":
    main()



from dataclasses import dataclass

from .decode import FileDecodingError, FileUnsupportedError


class OlDecodingError(FileDecodingError):
    """Base exception for texture files."""

    @property
    def prefix(self):
        return "Texture"


@dataclass
class OlUnsupportedFourcc(OlDecodingError, FileUnsupportedError):
    """Exception occurring when texture have unsupported or unknown fourcc."""

    fourcc: str

    def __str__(self):
        return f"{super().__str__()} has unsupported format: '{self.fourcc}'."



from typing import Optional

from scfile.file.base.decoder import DATA, OPENER, FileDecoder
from scfile.file.base.encoder import FileEncoder
from scfile.file.formats.dae import DaeEncoder
from scfile.file.formats.dds import DdsEncoder
from scfile.file.formats.mcsa import McsaDecoder
from scfile.file.formats.mic import MicDecoder
from scfile.file.formats.ms3d import Ms3dBinEncoder
from scfile.file.formats.ms3d_ascii import Ms3dAsciiEncoder
from scfile.file.formats.obj import ObjEncoder
from scfile.file.formats.ol import OlDecoder
from scfile.file.formats.ol.decoder import OlHdriDecoder
from scfile.file.formats.png import PngEncoder
from scfile.utils.types import PathLike

from .convert import convert


def converter(decoder: type[FileDecoder[OPENER, DATA]], encoder: type[FileEncoder[DATA]]):
    def decorator(_):
        def wrapper(source: PathLike, output: Optional[PathLike] = None, overwrite: bool = True):
            convert(source, output, decoder, encoder, overwrite)

        return wrapper

    return decorator


@converter(McsaDecoder, DaeEncoder)
def mcsa_to_dae(*args, **kwargs):
    """
    Converting model `.mcsa` file to `.dae`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `mcsa_to_dae("model.mcsa", "path/to/output")`
    """


@converter(McsaDecoder, ObjEncoder)
def mcsa_to_obj(*args, **kwargs):
    """
    Converting model `.mcsa` file to `.obj`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `mcsa_to_obj("model.mcsa", "path/to/output")`
    """


@converter(McsaDecoder, Ms3dBinEncoder)
def mcsa_to_ms3d(*args, **kwargs):
    """
    Converting model `.mcsa` file to `.ms3d`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `mcsa_to_ms3d("model.mcsa", "path/to/output")`
    """


@converter(McsaDecoder, Ms3dAsciiEncoder)
def mcsa_to_ms3d_ascii(*args, **kwargs):
    """
    Converting model `.mcsa` file to `.txt`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `mcsa_to_ms3d_ascii("model.mcsa", "path/to/output")`
    """


@converter(MicDecoder, PngEncoder)
def mic_to_png(*args, **kwargs):
    """
    Converting image `.mic` file to `.png`.

    Args:
        source: Path to `.mic` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `mic_to_png("image.mic", "path/to/output")`
    """


@converter(OlDecoder, DdsEncoder)
def ol_to_dds(*args, **kwargs):
    """
    Converting texture `.ol` file to `.dds`.

    Args:
        source: Path to `.ol` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `ol_to_dds("texture.ol", "path/to/output")`
    """


@converter(OlHdriDecoder, DdsEncoder)
def ol_hdri_to_dds(*args, **kwargs):
    """
    Converting hdri (sky) texture `.ol` file to `.dds`.

    Args:
        source: Path to `.ol` file.
        output (optional): Path to output directory. Defaults to source path with new suffix.
        overwrite (optional): Whether to overwrite output file if already exists. Defaults to True.

    Example:
        `ol_to_dds("texture.ol", "path/to/output")`
    """



from dataclasses import dataclass
from pathlib import Path

from scfile.utils.types import PathLike

from .base import ScFileException


@dataclass
class FileBasicError(ScFileException):
    """Basic files exception. Occurring with file processing."""

    file: PathLike

    @property
    def path(self):
        return Path(self.file)

    @property
    def prefix(self) -> str:
        return "File"

    def __str__(self):
        return f'{self.prefix} "{self.path.as_posix()}"'


@dataclass
class FileNotFound(FileBasicError):
    """Exception occurring when file not found or does not exists."""

    def __str__(self):
        return f"{super().__str__()} not found (not exists)."


@dataclass
class FileIsEmpty(FileBasicError):
    """Exception occurring when file is empty."""

    def __str__(self):
        return f"{super().__str__()} is empty."


@dataclass
class FileSuffixUnsupported(FileBasicError):
    """Exception occurring when file cannot be decoded or encoded."""

    def __str__(self):
        return f"{super().__str__()} with suffix '{self.path.suffix}' is unsupported."


@dataclass
class FileSignatureInvalid(FileBasicError):
    """Exception occurring when file signature does not match file type."""

    readed: int
    signature: int

    def __str__(self):
        return (
            f"{super().__str__()} has invalid signature - "
            f"{hex(self.readed)} != {hex(self.signature)}. "
            "(File suffix does not match file type)."
        )


@dataclass
class FileStructureInvalid(FileBasicError):
    """Exception occurring when file does not hold expected number of bytes."""

    pos: int

    def __str__(self):
        return f"{super().__str__()} has invalid structure. Current position: {self.pos}."



from dataclasses import dataclass

from .decode import FileDecodingError, FileParsingError, FileUnsupportedError


class McsaDecodingError(FileDecodingError):
    """Base exception for model files."""

    @property
    def prefix(self):
        return "Model"


@dataclass
class McsaCountsLimit(McsaDecodingError, FileParsingError):
    """Exception occurring when model counts is not read correctly."""

    counts: int

    def __str__(self):
        return f"{super().__str__()} has invalid structure. Model cannot have {self.counts:,} vertices/polygons."


@dataclass
class McsaUnknownLinkCount(McsaDecodingError, FileParsingError):
    """Exception occurring when model skeleton bones have unknown link count."""

    link_count: int

    def __str__(self):
        return f"{super().__str__()} has unknown bones link count: {self.link_count}."


@dataclass
class McsaUnsupportedVersion(McsaDecodingError, FileUnsupportedError):
    """Exception occurring when model version unsupported."""

    version: float

    def __str__(self):
        return f"{super().__str__()} has unsupported version: {self.version}."
